from .base import Model
from .helper import TypeFactory

MAGIC_PARSER_URL = "/v1/scan/text"


class MagicParser(object):
    @classmethod
    def parse(cls, text, session=None):
        session = session or Model._session

        payload = {'text': text}
        res = session.http_post(MAGIC_PARSER_URL, json=payload).json()

        for i in res['records']:
            yield TypeFactory.deserialize(i)



import logging

from .errors import SessionError

logger = logging.getLogger(__name__)

class_registry = {}

SYSREF_CLASS = "sysref"
SYSFACT_CLASS = "sysfact"
FACT_CLASS = "fact"
ANNOTATION_CLASS = "annotation"
REFERENCE_CLASS = "reference"


def sessionize(func):
    def wrapper(self, *args, **kwargs):
        session = None
        if func.__code__.co_argcount == 2:
            if len(args) == 1:
                session = args[0]
                args = ()
            else:
                session = kwargs.get('session')

        kwargs.setdefault('session', session or self._session)

        if kwargs.get('session') is None:
            raise SessionError("No session provided")

        return func(self, *args, **kwargs)
    return wrapper


def _register_class(key, value):
    class_registry[key] = value


def _resolve_classes(glob):
    for k, v in class_registry.items():
        class_registry[k] = glob[v]


def ClassFactory(name, argnames, baseClass):
    from .base import Fact, SysFact, Reference, SysRef

    class_map = {
        SYSREF_CLASS: SysRef,
        REFERENCE_CLASS: Reference,
        FACT_CLASS: Fact,
        SYSFACT_CLASS: SysFact
    }

    def __init__(self, *args, **kwargs):
        for key, value in kwargs.items():
            if key not in argnames:
                raise TypeError("Argument %s not valid for %s"
                                % (key, self.__class__.__name__))
        baseClass.__init__(self, *args, **kwargs)

    def get_class_name(name):
        return "".join([i.capitalize() for i in name.split('-')])

    baseClass = class_map[baseClass]
    newclass = type(get_class_name(name), (baseClass,), {"__init__": __init__})
    newclass._type = name
    newclass._phantom = True

    return newclass


class TypeFactory():
    """This class generates pyquo model objects from json"""
    class_argnames = {
        REFERENCE_CLASS: ('target', 'source', 'document', 'session'),
        SYSREF_CLASS: ('target', 'source', 'document', 'session'),
        FACT_CLASS: ('id', 'document', 'session'),
        SYSFACT_CLASS: ('id', 'document', 'session'),
        ANNOTATION_CLASS: ('fact', 'label', 'document', 'session')
    }

    @classmethod
    def _type_class(cls, type, argnames, baseClass):
        klass = class_registry.get(type)
        if klass is None:
            logger.warning('class %s not implemented', type)
            return ClassFactory(
                name=type,
                argnames=argnames,
                baseClass=baseClass
            )

        return klass

    @classmethod
    def _create_object(cls, type, baseClass):
        """This function allows generating unknown model classes on-the-fly.
        The class_argnames variable holds the definition of the constructor
        parameters for each model parent class type"""

        argnames = cls.class_argnames[baseClass]

        return cls._type_class(
            type=type,
            argnames=argnames,
            baseClass=baseClass
        )

    @classmethod
    def create_fact(cls, serialized):
        baseClass = serialized['class']

        factType = cls._create_object(serialized['type'], baseClass)
        obj = factType(id=serialized['id'])

        if 'document' in serialized:
            obj.document = serialized['document']

        return obj

    @classmethod
    def create_reference(cls, serialized):
        baseClass = serialized['class']
        target = serialized['target']
        source = serialized['source']

        # XXX remove once this is implemented serverside
        if target.get('class') is None:
            target['class'] = FACT_CLASS
        if source.get('class') is None:
            source['class'] = FACT_CLASS

        target = cls.create_fact(target)
        source = cls.create_fact(source)

        refType = cls._create_object(serialized['type'], baseClass)
        obj = refType(target=target, source=source)

        if 'document' in serialized:
            obj.document = serialized['document']

        if "index" in serialized:
            obj.index = serialized["index"]

        return obj

    @classmethod
    def create_annotation(cls, serialized):
        baseClass = serialized['class']

        # XXX remove once this is implemented serverside
        serialized['fact']['class'] = FACT_CLASS

        fact = cls.create_fact(serialized['fact'])
        annoType = cls._create_object(serialized['type'], baseClass)
        obj = annoType(fact=fact, label=serialized['label'])

        if 'document' in serialized:
            obj.document = serialized['document']

        return obj

    @classmethod
    def deserialize(cls, serialized):
        """This method deserializes a dictionary and returns a pyquo object"""
        baseClass = serialized.get('class', 'fact')

        if baseClass in (SYSREF_CLASS, REFERENCE_CLASS):
            return cls.create_reference(serialized)

        if baseClass in (SYSFACT_CLASS, FACT_CLASS):
            return cls.create_fact(serialized)

        if baseClass == ANNOTATION_CLASS:
            return cls.create_annotation(serialized)

        raise Exception('unknown base class %s', baseClass)



# Exceptions
class FetchError(Exception):
    """Raised when an error occurs during fetching"""
    status_code = None

    def __init__(self, error_msg, status_code=None):
        super(Exception, self).__init__(error_msg)
        self.status_code = status_code


class SessionError(Exception):
    """Raised when no session could be found"""
    pass


class ResultNotFound(FetchError):
    """Raised when the requested object is not found"""
    pass


class ValidationError(Exception):
    """Raised when the input validation fails"""
    pass


class RequiredError(Exception):
    """Raised when a required field is missing"""
    pass


class AuthenticationError(Exception):
    """Raised when the authentication failed"""
    pass



import os.path

from .session import FILE_UPLOAD
from .helper import _resolve_classes, sessionize
from .fields import Dict, List, String, Float, Integer, Flavor
from .base import Reference, SysRef, Fact, SysFact, Annotation


# Facts
class Certificate(Fact):
    _type = 'certificate'


class Email(Fact):
    _type = 'email'


class Envelope(Fact):
    _type = 'envelope'


class ExportTable(Fact):
    _type = 'export-table'


class File(Fact):
    _type = 'file'

    # md5 = String()
    # sha256 = String()
    # sha1 = String()

    # # XXX old format
    # mime_type = String(field='mime-type')
    # language_name = String()
    # language_id = Integer()
    # language_country = String()
    # type_name = String()
    # type_id = Integer()
    # tlsh = String()
    # magic = Dict()

    @staticmethod
    def _extract_filename(file_obj):
        name = getattr(file_obj, 'name', None)

        return os.path.basename(name) if name else None

    @classmethod
    @sessionize
    def upload(cls, file_obj, filename=None, session=None):
        """ This function uploads a file_obj to quolab

        :param file_obj: can be a file-like object such as BytesIO
             or a string
        :param str filename: Name of the file, if not provided it will be
        guessed based on file_obj.name

        :returns: An instance of pyquo.models.File
        """
        try:
            content = file_obj.read()
        except AttributeError:
            content = file_obj

        headers = {'Content-Type': 'application/octet-stream'}
        data = session.http_post(FILE_UPLOAD, data=content, headers=headers)
        f = cls(id=data.json()['records'][0]['id']).get(session)

        if filename is None:
            filename = cls._extract_filename(file_obj)

        if filename:
            KnownAs(f, filename).save(session=session)

        return f

    @sessionize
    def download(self, session=None):
        url = '/v1/file/{}'.format(self.id)
        return session.http_get(url)


class Function(Fact):
    _type = 'function'

    @sessionize
    def similar(self, session=None):
        path = '/v1/function/{}/similar'.format(self.id)
        response = session.http_get(path).json()
        similars = response.get('records', [])

        return (
            Function(id=match['id'], session=session)
            for match in similars
        )


class Hostname(Fact):
    _type = 'hostname'


class ImportTable(Fact):
    _type = 'import-table'


class IpAddress(Fact):
    _type = 'ip-address'

    version = Integer()


class Ipynb(Fact):
    _type = 'ipynb'

    notebook = Dict()


class Malware(Fact):
    _type = 'malware'


class MispBlob(Fact):
    _type = 'misp-blob'


class MispEvent(Fact):
    _type = 'misp-event'


class TTP(Fact):
    _type = 'ttp'


class Mutex(Fact):
    _type = 'mutex'


class Organization(Fact):
    _type = 'organization'


class Persona(Fact):
    _type = 'persona'


class PGP(Fact):
    _type = 'persona'

    key_id = String(field='key-id')


class Process(Fact):
    _type = 'process'


class RegistryKey(Fact):
    _type = 'registry-key'


class URL(Fact):
    _type = "url"

    netloc = String(read_only=True)
    scheme = String(read_only=True)
    path = String(read_only=True)
    query = String(read_only=True)
    hostname = String(read_only=True)


class Wallet(Fact):
    _type = 'wallet'


class YaraRule(Fact):
    _type = 'yara-rule'


class AutonomousSystem(Fact):
    _type = 'autonomous_system'


class Region(Fact):
    _type = 'region'


class TorDescriptor(Fact):
    _type = 'tor-descriptor'


# System Facts
class Case(SysFact):
    _type = "case"

    name = String(required=True)
    description = String(required=True, nullable=True)
    priority = Float(nullable=True)
    flavor = Flavor()
    type = String(required=True)

    created_at = Float(read_only=True, field='created-at')
    created_by = String(read_only=True, nullable=True, field='created-by')
    updated_at = Float(read_only=True, field='updated-at')
    contributors = List(read_only=True, type='Contributor')

    def __init__(self, id=None, **kwargs):
        super().__init__(id, **kwargs)

    @property
    def id(self):
        if self._id is None:
            self._id = '__new__'

        return self._id

    @id.setter
    def id(self, value):
        self._id = value


class User(SysFact):
    _type = 'user'

    @classmethod
    @sessionize
    def current(cls, session=None):
        user_id = session.current_user
        return User(id=user_id)


class Tag(SysFact):
    _type = 'tag'

    name = String(required=True)
    taxonomy = String(required=True)

    def __repr__(self):
        return '{._type}({.name})'.format(self, self)


class Endpoint(SysFact):
    _type = 'endpoint'

    @property
    def connector(self):
        connector = self.descendants(facts=Connector)
        if connector:
            return connector[0]


class Connector(SysFact):
    _type = 'connector'

    @property
    def endpoints(self):
        return self.ancestors(facts=Endpoint)


class Regulator(SysFact):
    _type = 'regulator'


class Source(SysFact):
    _type = 'source'


class Sink(SysFact):
    _type = 'sink'


class Task(SysFact):
    _type = 'task'


# References
class Accesses(Reference):
    _type = 'accesses'


class RelatesTo(Reference):
    _type = 'relates-to'


class Contains(Reference):
    _type = "contains"


class Creates(Reference):
    _type = 'creates'


class Delivered(Reference):
    _type = 'delivered'


class IdentifiedAs(Reference):
    _type = 'identified-as'


class IdentifiedBy(Reference):
    _type = 'identified-by'


class Loads(Reference):
    _type = 'loads'


class ReferredBy(Reference):
    _type = 'referred-by'


class SignedBy(Reference):
    _type = 'signed-by'


class ResolvedTo(Reference):
    _type = 'resolved-to'


class CommentedBy(SysRef):
    _type = 'commented-by'

    body = String(field='body')
    created_at = Float(read_only=True, field='created-at')


# SysRefs
class Queued(SysRef):
    _type = 'queued'

    queued_at = Float(read_only=True, field='queued-at')


class Scheduled(SysRef):
    _type = 'scheduled'

    scheduled_at = Float(read_only=True, field='scheduled-at')


class Executed(SysRef):
    _type = 'executed'

    executed_at = Float(read_only=True, field='executed-at')


class Canceled(SysRef):
    _type = 'canceled'

    canceled_at = Float(read_only=True, field='canceled-at')


class Failed(SysRef):
    _type = 'failed'

    failed_at = Float(read_only=True, field='failed-at')


class ObservedBy(SysRef):
    _type = 'observed-by'


class AssociatedWith(SysRef):
    _type = 'associated-with'


class Encases(SysRef):
    _type = 'encases'


class Tagged(SysRef):
    _type = 'tagged'


class Uses(SysRef):
    _type = 'uses'


class Implies(SysRef):
    _type = 'implies'


# Annotations
class Attribute(Annotation):
    _type = 'attribute'


class PublicSubkey(Annotation):
    _type = 'public-subkey'


class SecretSubkey(Annotation):
    _type = 'secret-subkey'


class Text(Annotation):
    _type = 'text'


class InterpretedAs(Annotation):
    _type = 'interpreted-as'


class KnownAs(Annotation):
    _type = 'known-as'


class Report(Annotation):
    _type = 'report'


_resolve_classes(globals())



from . import models         # noqa
from . import session        # noqa
from . import authenticator  # noqa

try:
    import os
    token = os.environ.get('QUOLAB_API_TOKEN')
    base_url = os.environ.get('QUOLAB_BASE_URL')

    if token and base_url:
        from . import base
        base.Model._session = session.Session(
            base_url=base_url,
            auth=authenticator.TokenAuthenticator(token)
        )
except KeyError:
    pass



import six
from .helper import TypeFactory
from collections import defaultdict

from .errors import ValidationError


# Validators
class Validator(object):
    """Callable base class to validate a given value"""
    def __call__(self, value):
        self.validate(value)


class StringValidator(Validator):
    """Validate strings"""
    @classmethod
    def validate(cls, value):
        valid = isinstance(value, str)

        if six.PY2 is True:
            valid |= isinstance(value, unicode)  # noqa

        if valid is False:
            raise ValidationError('expected a string for "{}"'.format(value))


class FloatValidator(Validator):
    """Validate floats"""
    @classmethod
    def validate(cls, value):
        if not isinstance(value, float) and not isinstance(value, int):
            raise ValidationError('expected an int for "{}"'.format(value))


class DictValidator(Validator):
    """Validate dicts"""
    @classmethod
    def validate(cls, value):
        if not isinstance(value, dict):
            raise ValidationError('execpted a dict for "{}"'.format(value))


class IntegerValidator(Validator):
    """Validate integers"""
    @classmethod
    def validate(cls, value):
        if not isinstance(value, int):
            raise ValidationError('expected an int for {} ({})'.format(
                value, type(value)))


class ChoiceValidator(Validator):
    """Validate that a flavor is either 'hypothesis', 'case' or folder'"""

    def __init__(self, choices):
        self._choices = choices

    def validate(self, value):
        if value not in self._choices:
            raise ValidationError('expected "{}" to be one of {}'.format(
                value, self._choices))


class ListValidator(Validator):
    """Validate that field is a List. If a type is given the validator will
    validate that all elements of the list are of the given type"""
    def __init__(self, type):
        self.type = type

    def validate(self, value):
        if not isinstance(value, list):
            raise ValidationError('expected a list for {}'.format(value))

        if not self.type:
            return

        for element in value:
            if not isinstance(element, self.type):
                raise ValidationError(
                    'expected list element to be of type {}'.format(self.type))


# Fields
class Field(object):
    """The Field class is the class of which all other fields are derived.
    It implements the __get__ and __set__ methods to store and read value
    information to and from the parent._data and parent._fields
    properities.
    """
    _toResolve = defaultdict(list)

    def __init__(self, field=None, required=False,
                 read_only=False, nullable=False, validators=()):
        """
        Constructor of Field

        :param field: if set to non-None the field will be used as a name
            mapping
        :param read_only: if set to True, the field will be set to read_only
            and never be passed to the import endpoint when creating the object
        :param nullable: if set to True, this field ensures that the value of
            this field can be None
        :param validators: Tuple of validator classes

        :Example:

        >>> created_at = String(read_only=True, field="created-at")
        """
        self.validators = validators
        self.nullable = nullable
        self.field = field
        self.read_only = read_only

        self.required = required

    def __get__(self, parent, owner):
        """Get a property using lazy fetching of the entire object if a value
        of the requested property is not cached in the _data field of the model
        instance. Otherwise the cached value is returned from _data. If lazy a
        lazy fetch is performed all cached values will be overwritten"""

        if parent is None:
            return self

        """lazy fetching"""
        if self not in parent._data:
            if parent._session:
                parent.get()

        return parent._data.get(self, Unset)

    def __set__(self, parent, value):
        """Store value of a given property into the _data field of the model
        instance and trigger validators"""
        if value is None and self.nullable is True:
            parent._data[self] = None
            return

        self.validate(value)
        parent._data[self] = value

    @classmethod
    def _resolve(cls, glob):
        """Allow resolves of class names as string references. Fields
        must implement the _resolve method"""
        for typename, lst in cls._toResolve.items():
            for ob in lst:
                ob._resolve(glob[typename])

    def validate(self, value):
        """Iterate over all validators and trigger them"""
        for validator in self.validators:
            validator(value)


# Field Types
class Any(Field):
    """Any field"""


class Unset(Field):
    """Unset field"""


class Integer(Field):
    """Integer field"""
    def __init__(self, *args, **kwargs):
        super(Integer, self).__init__(
            validators=(IntegerValidator(),), *args, **kwargs)


class String(Field):
    """String field"""
    def __init__(self, *args, **kwargs):
        super(String, self).__init__(
            validators=(StringValidator(),), *args, **kwargs)


class Flavor(Field):
    """Flavor field"""
    def __init__(self, *args, **kwargs):
        super(Flavor, self).__init__(
            validators=(ChoiceValidator(['hypothesis', 'case', 'folder']),),
            *args, **kwargs)


class Float(Field):
    """Float field"""
    def __init__(self, *args, **kwargs):
        super(Float, self).__init__(
            validators=(FloatValidator(),), *args, **kwargs)


class Relation(Field):
    """Relation field"""
    def __init__(self, *args, **kwargs):
        super(Relation, self).__init__(
            validators=(DictValidator(),), *args, **kwargs)

    def __set__(self, parent, value):
        from .models import Model

        if isinstance(value, Model):
            value = TypeFactory._type_class(value['type'])(
                id=value['id'],
                session=parent._session,
                **value.get('document', {})
            )
        parent._data[self] = value


class List(Field):
    """List field. If Type is not None the list will validate that all
    elements of the list are of the given type"""
    def __init__(self, type=None, *args, **kwargs):
        self.type = type

        if isinstance(type, str):
            self._toResolve[type].append(self)
            return super(List, self).__init__(*args, **kwargs)

        super(List, self).__init__(
            validators=(ListValidator(type),), *args, **kwargs)

    def __set__(self, parent, value):
        if self.type:
            value = [self.type(parent._session, **v) for v in value]

        super(List, self).__set__(parent, value)

    def _resolve(self, typ):
        self.validators = (ListValidator(typ),)
        self.type = typ


class Dict(Field):
    """Dict field"""
    def __init__(self, *args, **kwargs):
        super(Dict, self).__init__(
            validators=(DictValidator(),), *args, **kwargs)


Field._resolve(globals())



from .errors import AuthenticationError
from .session import AUTH_API


class Authenticator:
    """Authenticator parent class"""

    def __call__(self, session):
        self.authenticate(session)


class TokenAuthenticator(Authenticator):
    """Authenticate pyquo using a quolab Authenticator"""

    def __init__(self, token):
        self._token = token

    @classmethod
    def create_token(self, session, expires_in):
        """This method returns a session id"""

        info = {"expires-in": expires_in}
        res = session.http_post(AUTH_API, json=info).json()
        return res

    def authenticate(self, session):
        headers = {'Authorization': "Quoken {}".format(self._token)}

        if session.headers:
            session.headers.update(headers)
        else:
            session.headers = headers


class UserAuthenticator(Authenticator):
    """Authenticate pyquo using Basic Auth"""

    def __init__(self, username, password):
        self._username = username
        self._password = password

    def authenticate(self, session):
        data = {
            'username': self._username,
            'password': self._password
        }
        path = "/v1/auth/login"

        r = session.http_post(path, json=data)
        if r.status_code != 200:
            raise AuthenticationError(r.content)



# coding: utf8
from collections import Iterable

from .helper import (
    sessionize,
    TypeFactory, _register_class,
    FACT_CLASS, ANNOTATION_CLASS, SYSFACT_CLASS, REFERENCE_CLASS, SYSREF_CLASS
)
from .session import Filter
from .errors import ValidationError, RequiredError, ResultNotFound
from .fields import Field

from six import with_metaclass


class QuerySet(object):
    _parent = None

    def __init__(self, parent, refs=(), facts=(),
                 incoming=False, limit=None, session=None):
        self._parent = parent
        self._references = refs or (SysRef, Reference)
        self._facts = facts
        self._incoming = incoming
        self._limit = limit
        self._session = session

    def __call__(self, refs=(), facts=(), annotations=(),
                 incoming=None, limit=None, session=None):
        if not isinstance(facts, Iterable):
            facts = (facts,)
        self._facts = facts

        if not isinstance(refs, Iterable):
            refs = (refs,)
        self._references = refs or (SysRef, Reference)

        if incoming is not None:
            self._incoming = incoming

        if limit is not None:
            self._limit = limit

        if session is not None:
            self._session = session

        return self

    def query(self, key=None):
        results = self._parent._references(
            refs=self._references,
            facts=self._facts,
            incoming=self._incoming,
            limit=self._limit,
            session=self._session
        )

        return tuple(results)

    def __getitem__(self, key):
        results = self.query(key)
        return [self._target(result) for result in results][key]

    def __iter__(self):
        for item in self.query():
            yield self._target(item)

    def __len__(self):
        return len(self.query())

    def __repr__(self):
        result = list(self[:4])
        if len(result) > 3:
            result[3] = 'truncated...'
        return 'QuerySet{}'.format(str(result))


class ReferenceQuerySet(QuerySet):
    @property
    def facts(self):
        return FactQuerySet(
            parent=self._parent,
            refs=self._references,
            facts=self._facts,
            incoming=self._incoming,
            limit=self._limit,
            session=self._session
        )

    def _target(self, item):
        return item


class FactQuerySet(QuerySet):
    def _target(self, item):
        if self._incoming is False:
            return item.target
        else:
            return item.source


class AnnotationQuerySet(QuerySet):
    def __init__(self, *args, **kwargs):
        self._annotations = kwargs.get('annotations')
        super(AnnotationQuerySet, self).__init__(*args, **kwargs)

    def _target(self, item):
        return item

    def __call__(self, annotations=(), limit=None, session=None):
        if not isinstance(annotations, Iterable):
            annotations = (annotations,)
        self._annotations = annotations

        if limit is not None:
            self._limit = limit

        if session is not None:
            self._session = session

        return self

    def query(self, key=None):
        return tuple(self._parent._annotations(
            limit=self._limit,
            annotations=self._annotations,
            session=self._session
        ))


class ModelMetaClass(type):
    """This class adds the '_fields' attribute which corresponds to a
    mapping of {"property name": <field instance>} and a _fieldmap
    corresponding to a mapping between the actual property name and the
    name of the expected field. E.g. {'created_at': 'created-at'}
    """
    def __new__(cls, name, bases, attrs):
        if attrs.get('_type'):
            _register_class(attrs['_type'], name)

        attrs['_fields'] = dict()
        attrs['_fieldmap'] = dict()

        for key, field in attrs.items():
            if not isinstance(field, Field):
                continue

            if field.field:
                attrs['_fieldmap'][field.field] = key

            attrs['_fields'][key] = field

        klass = super(ModelMetaClass, cls).__new__(cls, name, bases, attrs)
        klass._filter = Filter(klass)
        return klass


class Model(with_metaclass(ModelMetaClass)):
    """All Model types are derived from this class."""
    _class = None
    _type = None
    _session = None

    def __init__(self, document=None, session=None, **kwargs):
        self._data = dict()
        for key, val in kwargs.items():
            if key in self._fields:
                self._data[self._fields[key]] = val
        if session is not None:
            self._session = session
        self.document = document

    @sessionize
    def get(self, session=None):
        """This method gets the object from the api using the session param
        or using the session provided by the instance. If no session is
        available this method raises a SessionError"""

        query = self.serialize
        records = session._query(query)
        if not records:
            raise ResultNotFound('Result not found for {}'.format(self))

        self.document = records[0]['document']

        return self

    @sessionize
    def save(self, session=None):
        """This method stores the object to the api using the session param
        or using the session provided by the instance. If no session is
        available this method raises a SessionError"""

        for key, field in self._fields.items():
            if field.required and field not in self._data:
                raise RequiredError('Required field "{}" missing {}'.format(
                    key, self))

            if field.read_only:
                continue

            if field in self._data:
                self.document[key] = self._data[field]

        serialized = self.serialize_with_document
        cl = serialized.pop('class')
        query = {self._class: [self.serialize_with_document]}
        res = session._import(query)
        res = res.get(cl)[0]
        res.setdefault('class', cl)
        return TypeFactory.deserialize(res)

    @sessionize
    def update(self, session=None, **patchfields):
        query = self.serialize
        query['patch'] = patchfields
        return session._patch(query)

    @sessionize
    def delete(self, session=None):
        """This method deletes an object using the api"""
        return session.remove({'query': self.serialize})

    @classmethod
    @sessionize
    def filter(self, target=None, source=None, fact=None,
               document=None, session=None, **kwargs):
        return self._filter(target, source, fact, document, session, **kwargs)

    def resolve_name(self, key):
        """This method resolves field names. E.g.

        >>> name = String(field='mapped-name')
        """
        return self._fieldmap.get(key, key)

    @property
    def document(self):
        return self._document

    @document.setter
    def document(self, document):
        if document is None:
            document = {}
        self._document = document

        for key, value in document.items():
            name = self.resolve_name(key)

            if name not in self._fields:
                field = Field(value)
                self._fields[name] = field
                setattr(self, name, field)

            try:
                setattr(self, name, value)
            except ValidationError:
                raise ValidationError("{} has no field {}={}".format(
                    self._type, name, value))

    @classmethod
    def deserialize(self, payload):
        return TypeFactory.deserialize(payload)

    @property
    def serialize_with_document(self):
        result = self.serialize
        if self.document is None:
            raise Exception('Document not fetched')

        result['document'] = self.document
        return result


# Base Models
class BaseFact(Model):
    """All Fact derived classes must have a unique id field and can load and
    save the field values to and from the api"""

    _id = None

    def __init__(self, id, document=None, session=None, **kwargs):
        super(BaseFact, self).__init__(
            session=session, document=document, **kwargs)

        self.id = id
        self.__references = ReferenceQuerySet(
            self, session=session or self._session)

    def get_references(self, ref, facts, incoming=True, session=None):
        """This function retrieves references"""

        query = {"class": ref._class}

        key, ikey = 'target', 'source'
        if not incoming:
            key, ikey = 'source', 'target'

        query[key] = self.serialize

        if ref._type:
            query['type'] = ref._type

        if facts:
            query[ikey] = []
            for fact in facts:
                q = {'class': fact._class, 'type': fact._type}
                query[ikey].append(q)

            if len(query[ikey]) == 1:
                query[ikey] = q

        return session._query(query)

    @sessionize
    def _references(self, limit, refs=(), facts=(),
                    incoming=False, session=None):
        for ref in refs:
            references = self.get_references(
                ref=ref, facts=facts, incoming=incoming, session=session)

            base = Fact if ref == Reference else SysFact

            for item in references:
                # XXX remove once this has been implementd serverside
                item['target'].setdefault('class', base._class)
                item['source'].setdefault('class', base._class)

                yield TypeFactory.deserialize(item)

    @property
    def references(self):
        return self.__references

    @property
    def descendants(self):
        return self.__references(incoming=False).facts

    @property
    def ancestors(self):
        return self.__references(incoming=True).facts

    @property
    def serialize(self):
        result = {
            'class': self._class,
            'type': self._type
        }

        if self.id:
            result['id'] = self.id

        return result

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    def __repr__(self):
        return '{._type}({.id})'.format(self, self)

    def __eq__(self, obj):
        return self.id == obj.id and \
                self._type == obj._type and \
                self._class == obj._class

    def __hash__(self):
        return hash(self.id) ^ hash(self._type) ^ hash(self._class)


class BaseReference(Model):
    def __init__(self, target, source, document=None, session=None):
        super(BaseReference, self).__init__(session=session, document=document)

        target._session = session or self._session
        source._session = session or self._session

        self.target = target
        self.source = source

    @property
    def serialize(self):
        return {
            'class': self._class,
            'type': self._type,
            'target': self.target.serialize,
            'source': self.source.serialize
        }

    def __repr__(self):
        return "{.source}→{._type}→{.target}".format(self, self, self)

    def __eq__(self, obj):
        try:
            # XXX check indexed field?
            return self.target == obj.target and \
                   self.source == obj.source
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self.source) ^ hash(self.target) ^ \
               hash(self._type) ^ hash(self._class)


class Fact(BaseFact):
    """Fact Model"""
    _class = FACT_CLASS

    def __init__(self, id, document=None, session=None, **kwargs):
        super(Fact, self).__init__(
            id, session=session, document=document, **kwargs)

        self.__annotations = AnnotationQuerySet(
            self, session=session or self._session)

    @property
    def annotations(self):
        return self.__annotations

    @sessionize
    def _annotations(self, annotations, limit, session=None):
        queries = []
        query = {"query": queries}

        if annotations:
            for annotation in annotations:
                queries.append({
                    "class": "annotation",
                    "type": annotation._type,
                    "fact": self.serialize
                })
        else:
            queries.append({"class": "annotation", "fact": self.serialize})

        if limit is not None:
            query['limit'] = limit

        for item in session._query(query):
            yield TypeFactory.deserialize(item)

    def __repr__(self):
        return '{._type}({.id})'.format(self, self)


class Annotation(Model):
    """Annotation Model"""
    _class = ANNOTATION_CLASS

    @property
    def serialize(self):
        result = {
            'type': self._type,
            'class': self._class,
            'label': self.label
        }
        if self.fact:
            result['fact'] = self.fact.serialize

        return result

    def __init__(self, fact, label, document=None, session=None):
        super(Annotation, self).__init__(session=session, document=document)

        self.label = label
        self.fact = fact

    def __repr__(self):
        return "{.fact}♪{._type}({.label})".format(self, self, self)


class SysFact(BaseFact):
    """SysFact Model"""
    _class = SYSFACT_CLASS


class Reference(BaseReference):
    """Reference Model"""
    _class = REFERENCE_CLASS


class SysRef(BaseReference):
    """SysRef Model"""
    _class = SYSREF_CLASS



from requests import Session
import json

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from .errors import FetchError
from .helper import TypeFactory

import logging

logger = logging.getLogger(__name__)


CATALOG_QUERY = "/v1/catalog/query"
CATALOG_IMPORT = "/v1/catalog/import"
AUTH_API = "/v1/auth/api"
AUTH_LOGIN = "/v1/auth/login"
FILE_UPLOAD = '/v1/file'
FILE_DOWNLOAD = FILE_UPLOAD + '/{}'


def expect(*codes):
    def wrap(func, *args, **kwargs):
        def wrapper(*args, **kwargs):
            res = func(*args, **kwargs)
            if res.status_code not in codes:
                raise FetchError("unexpected http code <{}> {}".format(
                    res.status_code, res.content), res.status_code)
            return res
        return wrapper
    return wrap


def logme(func):
    def wrap(self, path, *args, **kwargs):
        name = func.__code__.co_name.upper()
        logger.debug('%s %s data: %s', name.upper(), path, kwargs)
        res = func(self, path, *args, **kwargs)
        try:
            resp = res.json()
        except Exception:
            resp = res.content
        logger.debug('%d: Response %s', res.status_code, resp)
        return res
    return wrap


def set_global_session(session):
    from .base import Model
    Model._session = session


class Session(Session):
    """ This class maintains the session information for the current
        Connection with the QuoLab server.
        An Instance of this class is required when performing operations
        against the API

        >>> s = Session("http://qlab01-dev.app.quo:9090/v1/", None)
        >>> s.get(Case, "2ee9b31253f44e6cb1d40ff7af333b4f")

    """
    def __init__(self, base_url, verify=True, auth=None, global_session=False,
                 *args, **kwargs):
        super(Session, self).__init__(*args, **kwargs)
        self.url = base_url
        self.verify = verify
        if global_session is True:
            set_global_session(self)

        if auth:
            auth(self)

    @logme
    def http_post(self, path, data=None, json=None, headers={}):
        url = urljoin(self.url, path)
        res = self.post(url, data=data, json=json,
                        headers=headers, verify=self.verify)
        return res

    @logme
    def http_patch(self, path, data=None, json=None, headers={}):
        url = urljoin(self.url, path)
        res = self.patch(url, data=data, json=json,
                         headers=headers, verify=self.verify)
        return res

    @logme
    def http_delete(self, path, data=None, json=None, headers={}):
        url = urljoin(self.url, path)
        res = self.delete(url, data=data, json=json,
                          headers=headers, verify=self.erify)
        return res

    @logme
    def http_get(self, path):
        url = urljoin(self.url, path)
        res = self.get(url, verify=self.verify)
        return res

    def remove(self, query):
        @expect(200)
        def q(self, query):
            return self.http_delete(CATALOG_QUERY, json=query)
        return q(self, query).json()

    def _query(self, query):
        @expect(200)
        def q(self, query):
            if isinstance(query, str):
                query = json.dumps(query)

            return self.http_post(CATALOG_QUERY, json=query)
        return q(self, query).json()['records']

    def _import(self, query):
        @expect(200)
        def q(self, query):
            return self.http_post(CATALOG_IMPORT, json=query)
        return q(self, query).json()

    def _patch(self, query):
        @expect(200)
        def q(self, query):
            return self.http_patch(CATALOG_QUERY, json=query)
        return q(self, query).json()

    @property
    def current_user(self):
        @expect(200)
        def q(self):
            return self.http_get(AUTH_LOGIN)
        return q(self).json()['user']['id']


class Filter():
    def __init__(self, parent):
        self._type = parent._type
        self._class = parent._class

    def __call__(self, target=None, source=None, fact=None,
                 document=None, session=None, **kwargs):
        if target:
            kwargs['target'] = target.serialize
        if source:
            kwargs['source'] = source.serialize
        if fact:
            kwargs['fact'] = [f.serialize for f in fact]

        results = self.query(
            session=session,
            _type=self._type,
            _class=self._class,
            document=document,
            **kwargs
        )

        return [TypeFactory.deserialize(r) for r in results]

    def query(self, session, _type, _class, document=None, **kwargs):
        query = dict(kwargs)
        query.update({
            'type': _type,
            'class': _class,
        })

        if isinstance(document, dict):
            query['document'] = document

        return session._query(query)


class Query():
    """This class allows issuing raw queries to the API"""
    @classmethod
    def generate(cls, query, session):
        """This method returns serialized pyquo objects"""
        for i in cls.execute(query, session):
            yield TypeFactory.deserialize(i)

    @classmethod
    def execute(cls, query, session):
        """This method returns the raw query request"""
        return session._query(query)



from setuptools import setup
import pathlib

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(name='pyquo',
      version='20.1.5',
      description='Python QuoLab REST Client',
      long_description=README,
      long_description_content_type='text/markdown',
      author='QuoLab Technologies',
      author_email='curious@quolab.com',
      packages=['pyquo'],
      url='https://github.com/quolab/pyquo',
      install_requires=[
          'certifi==2018.8.24',
          'chardet==3.0.4',
          'idna==2.6',
          'requests==2.21.0',
          'six==1.12.0',
          'urllib3==1.24.2'
      ])
